from tokenize import Bracket
import torchvision
import torch
from torch.optim.swa_utils import AveragedModel, SWALR, update_bn
import numpy as np
import pycls.datasets.loader as imagenet_loader
from pycls.core.model_builder import build_model
import pycls.core.losses as losses
import pycls.core.model_builder as model_builder
import pycls.utils.distributed as du
import pycls.utils.checkpoint as cu
import pycls.utils.net as nu
import pycls.core.optimizer as optim
import pycls.utils.benchmark as bu
import pycls.utils.metrics as mu
import pycls.utils.logging as lu
from pycls.utils.meters import TestMeter
from pycls.utils.meters import TrainMeter
from pycls.utils.meters import ValMeter
from tqdm import tqdm
import copy
import os
from glob import glob
# from tools.train_al_swatorch import test_epoch

logger = lu.get_logger(__name__)

def train_swa_epoch(train_loader, model, swa_model, swa_freq, swa_scheduler, loss_fun, optimizer, train_meter, cur_epoch, cfg):
    if cfg.NUM_GPUS>1:
        if du.is_master_proc(cfg): logger.info('This should be deinitely if du.is_master_proc(cfg): logger.info in DDP -- setting cur-epoch in train loader')
        train_loader.sampler.set_epoch(cur_epoch)

    # No need to Update the learning rate as we use fixed LR for implementing SWA
    lr = optim.get_epoch_lr(cfg, cur_epoch)
    # if cfg.OPTIM.TYPE == "sgd":
    #     optim.set_lr(optimizer, lr)

    # model.train()
    train_meter.iter_tic() #This basically notes the start time in timer class defined in utils/timer.py

    len_train_loader = len(train_loader)
    print('len train loader: ', len_train_loader)
    for cur_iter, (inputs, labels) in enumerate(train_loader):

        #ensuring that inputs are floatTensor as model weights are
        inputs = inputs.type(torch.cuda.FloatTensor)

        inputs, labels = inputs.cuda(), labels.cuda(non_blocking=True)

        # Perform the forward pass
        preds = model(inputs)
        # Compute the loss
        loss = loss_fun(preds, labels)
        # Perform the backward pass
        optimizer.zero_grad()
        loss.backward()

        # Update the parametersSWA
        optimizer.step()
        # Compute the errors

        # add model weights to running average of weights
        if cur_iter!=0 and cur_iter%swa_freq==0:
            swa_model.update_parameters(model)
            swa_scheduler.step()
        
        top1_err, top5_err = mu.topk_errors(preds, labels, [1, 5])
        
        # Combine the stats across the GPUs
        if cfg.NUM_GPUS > 1:
            #Average error and losses across GPUs
            #Also this this calls wait method on reductions so we are ensured
            #to obtain synchronized results
            loss, top1_err = du.scaled_all_reduce(cfg, 
                [loss, top1_err]
            )
        
        # Copy the stats from GPU to CPU (sync point)
        loss, top1_err = loss.item(), top1_err.item()
        
        #Compute the difference in time now from start time initialized just before this for loop.
        train_meter.iter_toc()

        train_meter.update_stats(top1_err=top1_err, loss=loss, \
            lr=lr, mb_size=inputs.size(0) * cfg.NUM_GPUS)

        train_meter.log_iter_stats(cur_epoch, cur_iter)
        train_meter.iter_tic()

        print(f'Iteration [{cur_iter}/{len_train_loader}]')
        

    # Log epoch stats
    train_meter.log_epoch_stats(cur_epoch)
    train_meter.reset()

    return loss, swa_model, swa_scheduler

@torch.no_grad()
def test_epoch(cfg, test_loader, model, test_meter, cur_epoch):
    """Evaluates the model on the test set."""

    # Enable eval mode
    model.eval()
    test_meter.iter_tic()

    misclassifications = 0.
    totalSamples = 0.
    for cur_iter, (inputs, labels) in enumerate(test_loader):
        # Transfer the data to the current GPU device
        inputs, labels = inputs.cuda(), labels.cuda(non_blocking=True)
        inputs = inputs.type(torch.cuda.FloatTensor)

        # Compute the predictions
        preds = model(inputs)
        # Compute the errors
        top1_err, top5_err = mu.topk_errors(preds, labels, [1, 5])
        # Combine the errors across the GPUs
        if cfg.NUM_GPUS > 1:
            top1_err = du.scaled_all_reduce(cfg, [top1_err])
            #as above returns a list
            top1_err = top1_err[0]
        # Copy the errors from GPU to CPU (sync point)
        top1_err = top1_err.item()

        #Multiply by Number of GPU's as top1_err is scaled by 1/Num_GPUs
        misclassifications += top1_err * inputs.size(0) * cfg.NUM_GPUS
        totalSamples += inputs.size(0)*cfg.NUM_GPUS

        test_meter.iter_toc()
        # Update and log stats
        test_meter.update_stats(
            top1_err=top1_err, mb_size=inputs.size(0) * cfg.NUM_GPUS
        )
        test_meter.log_iter_stats(cur_epoch, cur_iter)
        test_meter.iter_tic()

    # Log epoch stats
    test_meter.log_epoch_stats(cur_epoch)
    test_meter.reset()

    return misclassifications/totalSamples

def swa_train_model(argListSWA):
    # Determine the GPU used by the current process
    cur_device = torch.cuda.current_device()
    start_epoch = 0
    args, load_model_path, lSetPath, uSetPath, save_path, trainDataset, noAugDataset, temp_cfg = argListSWA

    cfg = copy.deepcopy(temp_cfg)
    save_path = os.path.abspath(os.path.join(save_path, os.pardir))
    # if du.is_master_proc(cfg): logger.info('old_cfg.OUT_DIR: ', cfg.OUT_DIR)
    cfg.LOG_PERIOD = 1
    cfg.OUT_DIR = save_path
    swa_freq = args.swa_freq
    # if du.is_master_proc(cfg): logger.info('new cfg.OUT_DIR: ', cfg.OUT_DIR)
    
    isDistributedFlag = False
    if cfg.NUM_GPUS>1:
        isDistributedFlag = True
    
    # get model and optimizer
    # set isDistributed Flag=False here but then load model from path and then instantiate DDP
    model = model_builder.build_model(cfg, active_sampling=cfg.ACTIVE_LEARNING.ACTIVATE, isDistributed=isDistributedFlag)

    # load model
    if du.is_master_proc(cfg): logger.info(f"Model loading from path: {load_model_path}")
    
    cu.load_checkpoint(cfg, load_model_path, model)        

    loss_fun = losses.get_loss_fun()

    if du.is_master_proc(cfg):
        logger.info('=================================')
        logger.info('Loaded initial weights from: {}'.format(load_model_path))
        logger.info('Base LR: {}'.format(args.lr))
        logger.info('=================================')

    if cfg.OPTIM.TYPE.lower() == 'sgd': 
        optimizer = torch.optim.SGD(model.parameters(), lr=args.lr, weight_decay=args.wt_decay)
    else: 
        optimizer = torch.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.wt_decay)
    
    for state in optimizer.state.values():
        for key, val in state.items():
            #if isinstance(v, torch.Tensor):
            if torch.is_tensor(val):
                state[key] = val.cuda(cur_device)

    # Compute precise time
    if start_epoch == 0 and cfg.PREC_TIME.ENABLED:
        if du.is_master_proc(cfg): logger.info('Computing precise time...')
        bu.compute_precise_time(model, loss_fun)
        nu.reset_bn_stats(model)
    
    # get data loaders
    if cfg.TRAIN.DATASET == "IMAGENET":

        if du.is_master_proc(cfg): logger.info("==========================")
        if du.is_master_proc(cfg): logger.info("Load imagenet dataset")
        if du.is_master_proc(cfg): logger.info("==========================")
        if du.is_master_proc(cfg): logger.info('lSetPath arg passed: ', lSetPath)
        if du.is_master_proc(cfg): logger.info('uSetPath arg passed: ', uSetPath)
        cfg.RANDAUG.ACTIVATE = False
        train_loader, valSetLoader = imagenet_loader.get_data_loaders(cfg)
        test_loader = imagenet_loader.construct_test_loader(cfg)
    
    # Create meters
    train_meter = TrainMeter(len(train_loader), cfg)
    val_meter   = ValMeter(len(valSetLoader)  , cfg)
    test_meter  = TestMeter(len(test_loader)  , cfg)

    # Perform the training loop
    if du.is_master_proc(cfg): logger.info("Len(train_loader): {}".format(len(train_loader)))
    if du.is_master_proc(cfg): logger.info('Start epoch: {}'.format(start_epoch + 1))
    val_set_acc = 0.

    #set things for running swa
    model.train()
    
    # keep track of running average over weights -- this will be our final model
    if du.is_master_proc(cfg): logger.info('Constructing SWA Averaged Model...')
    swa_model = AveragedModel(model)
    if du.is_master_proc(cfg): logger.info('Done..!!')

    # scheduler wrapping the optimizer
    # custom choice as initial lr can disturb the weights drastically so anneal the lr for few epochs (see anneal_epochs param)
    if du.is_master_proc(cfg): logger.info('Constructing SWA Scheduler wrapping the optimizer...')
    swa_scheduler = SWALR(optimizer,anneal_strategy="linear", anneal_epochs=5, swa_lr=args.swa_lr)
    if du.is_master_proc(cfg): logger.info('Done..!!')
    
    for cur_epoch in range(start_epoch, args.swa_epochs):

        ## TRAIN SWA FOR ONE EPOCH
        swa_train_loss, swa_model, swa_scheduler = train_swa_epoch(train_loader, model, swa_model, swa_freq, \
                        swa_scheduler, loss_fun, optimizer, train_meter, cur_epoch, cfg)
        # swa_train_loss = train_swa_epoch(train_loader, model, loss_fun, optimizer, train_meter, cur_epoch, cfg)

        # Compute precise BN stats
        if cfg.BN.USE_PRECISE_STATS:
            nu.compute_precise_bn_stats(model, train_loader)

        if du.is_master_proc(cfg): logger.info("[SWA Distrib] Train Loss: {}".format(swa_train_loss))

    # update bn stats
    print('Updating BN Stats!!')
    update_bn(train_loader, swa_model)
    print('Done..!')

    # compute the validation accuracy
    val_set_err = test_epoch(cfg, valSetLoader, model, val_meter, args.swa_epochs)
    val_set_acc = 100. - val_set_err
    if du.is_master_proc(cfg): logger.info('SWA model accuracy: ', val_set_acc)

    print('~~~ cfg.OUT_DIR: ', cfg.OUT_DIR)
    checkpoint_file = cu.named_save_checkpoint(cfg,"[SWA]valSet_acc_"+str(val_set_acc), model, optimizer, args.swa_epochs)
    
    logger.info('Wrote checkpoint to: {}'.format(checkpoint_file))

def swa_single_proc_train(args, load_model_path, lSetPath, uSetPath, save_path, trainDataset, noAugDataset, cfg): #(argListSWA)
    # args, load_model_path, lSetPath, uSetPath, save_path, trainDataset, noAugDataset, cfg = argListSWA
    argListSWA = [args, load_model_path, lSetPath, uSetPath, save_path, trainDataset, noAugDataset, cfg]

    np.random.seed(cfg.RNG_SEED)
    torch.manual_seed(cfg.RNG_SEED)
    # Configure the CUDNN backend
    torch.backends.cudnn.benchmark = cfg.CUDNN.BENCHMARK

    # Train the model
    swa_train_model(argListSWA)